from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from django.core.files.base import ContentFile
from django.utils.module_loading import import_string

from ninja_simple_jwt.settings import ninja_simple_jwt_settings


def make_and_save_key_pair() -> tuple[str, str]:
    jwt_key_storage = import_string(ninja_simple_jwt_settings.JWT_KEY_STORAGE)
    pem_private_key, pem_public_key = make_keys()
    private_key_path = jwt_key_storage.save(
        name=ninja_simple_jwt_settings.JWT_PRIVATE_KEY_PATH, content=ContentFile(pem_private_key)
    )
    public_key_path = jwt_key_storage.save(
        name=ninja_simple_jwt_settings.JWT_PUBLIC_KEY_PATH, content=ContentFile(pem_public_key)
    )
    return private_key_path, public_key_path


def make_keys() -> tuple[bytes, bytes]:
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=2048)

    pem_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption(),
    )

    pem_public_key = private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM, format=serialization.PublicFormat.SubjectPublicKeyInfo
    )

    return pem_private_key, pem_public_key



from django.utils.functional import classproperty
from django.utils.module_loading import import_string

from ninja_simple_jwt.settings import ninja_simple_jwt_settings


class InMemoryJwtKeyPair:
    _private_key = None
    _public_key = None

    @classproperty
    def private_key(self) -> bytes:
        if self._private_key is None:
            self._private_key = self._get_private_jwt_key()
        return self._private_key

    @classproperty
    def public_key(self) -> bytes:
        if self._public_key is None:
            self._public_key = self._get_public_jwt_key()
        return self._public_key

    @staticmethod
    def _get_private_jwt_key() -> bytes:
        jwt_key_storage = import_string(ninja_simple_jwt_settings.JWT_KEY_STORAGE)
        with jwt_key_storage.open(ninja_simple_jwt_settings.JWT_PRIVATE_KEY_PATH) as f:
            return f.read()

    @staticmethod
    def _get_public_jwt_key() -> bytes:
        jwt_key_storage = import_string(ninja_simple_jwt_settings.JWT_KEY_STORAGE)
        with jwt_key_storage.open(ninja_simple_jwt_settings.JWT_PUBLIC_KEY_PATH) as f:
            return f.read()



from django.http import HttpRequest
from jwt import PyJWTError
from ninja.errors import AuthenticationError
from ninja.security import HttpBearer
from ninja.security.http import DecodeError

from ninja_simple_jwt.jwt.token_operations import TokenTypes, decode_token


class HttpJwtAuth(HttpBearer):
    def authenticate(self, request: HttpRequest, token: str) -> bool:
        token = self.decode_authorization(request.headers["Authorization"])

        try:
            access_token = decode_token(token, token_type=TokenTypes.ACCESS, verify=True)
        except PyJWTError as e:
            raise AuthenticationError(e)

        setattr(request.user, "id", access_token["user_id"])
        setattr(request.user, "username", access_token["username"])

        return True

    def decode_authorization(self, value: str) -> str:
        parts = value.split(" ")
        if len(parts) != 2 or parts[0].lower() != "bearer":
            raise DecodeError("Invalid Authorization header")

        token = parts[1]
        return token



import time
from datetime import datetime
from enum import Enum
from typing import Any, Tuple, TypedDict
from uuid import uuid4

import jwt
from django.contrib.auth.models import AbstractBaseUser
from django.utils import timezone
from jwt import ExpiredSignatureError, InvalidKeyError, InvalidTokenError
from pydantic import UUID4

from ninja_simple_jwt.jwt.key_retrieval import InMemoryJwtKeyPair
from ninja_simple_jwt.settings import ninja_simple_jwt_settings


class TokenTypes(str, Enum):
    ACCESS = "access"
    REFRESH = "refresh"


class TokenPayload(TypedDict):
    user_id: str
    username: str


class DecodedTokenPayload(TokenPayload):
    token_type: TokenTypes
    jti: UUID4
    exp: int
    iat: int


def get_refresh_token_for_user(user: AbstractBaseUser) -> Tuple[str, dict]:
    return encode_token({"username": user.get_username(), "user_id": user.pk}, TokenTypes.REFRESH)


def get_access_token(user_id: str, user_name: str) -> Tuple[str, dict]:
    return encode_token({"username": user_name, "user_id": user_id}, TokenTypes.ACCESS)


def get_access_token_from_refresh_token(refresh_token: str) -> Tuple[str, dict]:
    decoded = decode_token(refresh_token, token_type=TokenTypes.REFRESH, verify=True)
    return get_access_token(user_id=decoded["user_id"], user_name=decoded["username"])


def encode_token(payload: TokenPayload, token_type: TokenTypes, **additional_headers: Any) -> Tuple[str, dict]:
    now = timezone.now()
    if token_type == TokenTypes.REFRESH:
        expiry = now + ninja_simple_jwt_settings.JWT_REFRESH_TOKEN_LIFETIME
    else:
        expiry = now + ninja_simple_jwt_settings.JWT_ACCESS_TOKEN_LIFETIME

    payload_data = {
        **payload,
        "user_id": str(payload["user_id"]),
        "jti": uuid4().hex,
        "exp": int(time.mktime(expiry.timetuple())),
        "iat": int(time.mktime(now.timetuple())),
        "token_type": token_type,
    }

    return (
        jwt.encode(payload_data, InMemoryJwtKeyPair.private_key, algorithm="RS256", headers=additional_headers),
        payload_data,
    )


def decode_token(token: str, token_type: TokenTypes, verify: bool = True) -> DecodedTokenPayload:
    if verify is True:
        decoded = jwt.decode(token, InMemoryJwtKeyPair.public_key, algorithms=["RS256"])
        _verify_exp(decoded)
        _verify_jti(decoded)
        _verify_token_type(decoded, token_type)
    else:
        decoded = jwt.get_unverified_header(token)
    return decoded


def _verify_exp(payload: dict) -> None:
    now = timezone.now()
    token_expiry_unix_time = payload["exp"]
    token_expiry = timezone.make_aware(datetime.fromtimestamp(token_expiry_unix_time))
    if now >= token_expiry:
        raise ExpiredSignatureError("JWT has expired.")


def _verify_jti(payload: dict) -> None:
    if "jti" not in payload:
        raise InvalidKeyError("Invalid jti claim in JWT.")


def _verify_token_type(payload: dict, token_type: TokenTypes) -> None:
    if "token_type" not in payload:
        raise InvalidKeyError("Missing token type in JWT.")
    if payload["token_type"] != token_type:
        raise InvalidTokenError("Incorrect token type in JWT.")



from typing import IO

from django.core.files.storage import Storage


class LocalDiskKeyStorage(Storage):
    def _save(self, name: str, content: IO) -> str:
        with open(name, "w", encoding="utf-8") as f:
            f.write(content.read().decode())
        return name

    def exists(self, name: str) -> bool:
        """Always overwrite old keys with same filename."""
        return False

    def _open(self, name: str, mode: str = "rb") -> IO:
        return open(name, mode)  # pylint: disable=unspecified-encoding


local_disk_key_storage = LocalDiskKeyStorage()



from datetime import timedelta
from typing import Any, Optional, TypedDict

from django.conf import settings
from django.test.signals import setting_changed
from typing_extensions import NotRequired

USER_SETTINGS = getattr(settings, "NINJA_SIMPLE_JWT", None)


class NinjaSimpleJwtSettingsDict(TypedDict):
    JWT_KEY_STORAGE: NotRequired[str]
    JWT_PRIVATE_KEY_PATH: NotRequired[str]
    JWT_PUBLIC_KEY_PATH: NotRequired[str]
    JWT_REFRESH_COOKIE_NAME: NotRequired[str]
    JWT_REFRESH_TOKEN_LIFETIME: NotRequired[timedelta]
    JWT_ACCESS_TOKEN_LIFETIME: NotRequired[timedelta]
    WEB_REFRESH_COOKIE_SECURE: NotRequired[bool]
    WEB_REFRESH_COOKIE_HTTP_ONLY: NotRequired[bool]
    WEB_REFRESH_COOKIE_SAME_SITE_POLICY: NotRequired[str]
    WEB_REFRESH_COOKIE_PATH: NotRequired[str]


DEFAULTS: NinjaSimpleJwtSettingsDict = {
    "JWT_KEY_STORAGE": "ninja_simple_jwt.jwt.key_store.local_disk_key_storage",
    "JWT_PRIVATE_KEY_PATH": "jwt-signing.pem",
    "JWT_PUBLIC_KEY_PATH": "jwt-signing.pub",
    "JWT_REFRESH_COOKIE_NAME": "refresh",
    "JWT_REFRESH_TOKEN_LIFETIME": timedelta(days=30),
    "JWT_ACCESS_TOKEN_LIFETIME": timedelta(minutes=15),
    "WEB_REFRESH_COOKIE_SECURE": not settings.DEBUG,
    "WEB_REFRESH_COOKIE_HTTP_ONLY": True,
    "WEB_REFRESH_COOKIE_SAME_SITE_POLICY": "Strict",
    "WEB_REFRESH_COOKIE_PATH": "/api/auth/web/token-refresh",
}

EMPTY_SETTINGS: NinjaSimpleJwtSettingsDict = {}


class NinjaSimpleJwtSettings:
    def __init__(
        self,
        user_settings: Optional[NinjaSimpleJwtSettingsDict] = None,
        defaults: Optional[NinjaSimpleJwtSettingsDict] = None,
    ) -> None:
        self._user_settings = user_settings or EMPTY_SETTINGS
        self.defaults = defaults or DEFAULTS
        self._cached_attrs: set = set()

    @property
    def user_settings(self) -> NinjaSimpleJwtSettingsDict:
        if not hasattr(self, "_user_settings"):
            self._user_settings = getattr(settings, "NINJA_SIMPLE_JWT", EMPTY_SETTINGS)
        return self._user_settings

    def __getattr__(self, attr: str) -> Any:
        # check the setting is accepted
        if attr not in self.defaults:
            raise AttributeError(f"Invalid NINJA_SIMPLE_JWT setting: {attr}")

        try:
            val = self.user_settings[attr]  # type: ignore
        except KeyError:
            val = self.defaults[attr]  # type: ignore

        self._cached_attrs.add(attr)
        setattr(self, attr, val)
        return val

    def reload(self) -> None:
        for attr in self._cached_attrs:
            delattr(self, attr)
        self._cached_attrs.clear()
        if hasattr(self, "_user_settings"):
            delattr(self, "_user_settings")


ninja_simple_jwt_settings = NinjaSimpleJwtSettings(USER_SETTINGS, DEFAULTS)


def reload_drf_stripe_settings(*args: Any, **kwargs: Any) -> None:
    setting = kwargs["setting"]
    if setting == "NINJA_SIMPLE_JWT":
        ninja_simple_jwt_settings.reload()


setting_changed.connect(reload_drf_stripe_settings)



from datetime import UTC, datetime

from django.contrib.auth import authenticate
from django.http import HttpRequest, HttpResponse
from jwt.exceptions import PyJWTError
from ninja import Router
from ninja.errors import AuthenticationError

from ninja_simple_jwt.auth.views.schemas import (
    MobileSignInResponse,
    MobileTokenRefreshRequest,
    MobileTokenRefreshResponse,
    SignInRequest,
    WebSignInResponse,
)
from ninja_simple_jwt.jwt.token_operations import (
    get_access_token,
    get_access_token_from_refresh_token,
    get_refresh_token_for_user,
)
from ninja_simple_jwt.settings import ninja_simple_jwt_settings

mobile_auth_router = Router()
web_auth_router = Router()


@mobile_auth_router.post("/sign-in", response=MobileSignInResponse)
def mobile_sign_in(request: HttpRequest, payload: SignInRequest) -> dict:
    payload_data = payload.dict()
    user = authenticate(username=payload_data["username"], password=payload_data["password"])
    if user is None:
        raise AuthenticationError()
    refresh_token, _ = get_refresh_token_for_user(user)
    access_token, _ = get_access_token(str(user.pk), user.get_username())
    return {"refresh": refresh_token, "access": access_token}


@mobile_auth_router.post("/token-refresh", response=MobileTokenRefreshResponse)
def mobile_token_refresh(request: HttpRequest, payload: MobileTokenRefreshRequest) -> dict:
    payload_data = payload.dict()
    try:
        access_token, _ = get_access_token_from_refresh_token(payload_data["refresh"])
    except PyJWTError:
        raise AuthenticationError()

    return {"access": access_token}


@web_auth_router.post("/sign-in", response=WebSignInResponse)
def web_sign_in(request: HttpRequest, payload: SignInRequest, response: HttpResponse) -> dict:
    payload_data = payload.dict()
    user = authenticate(username=payload_data["username"], password=payload_data["password"])
    if user is None:
        raise AuthenticationError()
    refresh_token, refresh_token_payload = get_refresh_token_for_user(user)
    access_token, _ = get_access_token(str(user.pk), user.get_username())
    response.set_cookie(
        key=ninja_simple_jwt_settings.JWT_REFRESH_COOKIE_NAME,
        value=refresh_token,
        expires=datetime.fromtimestamp(refresh_token_payload["exp"], UTC),
        httponly=ninja_simple_jwt_settings.WEB_REFRESH_COOKIE_HTTP_ONLY,
        samesite=ninja_simple_jwt_settings.WEB_REFRESH_COOKIE_SAME_SITE_POLICY,
        secure=ninja_simple_jwt_settings.WEB_REFRESH_COOKIE_SECURE,
        path=ninja_simple_jwt_settings.WEB_REFRESH_COOKIE_PATH,
    )
    return {"access": access_token}


@web_auth_router.post("/token-refresh", response=WebSignInResponse)
def web_token_refresh(request: HttpRequest) -> dict:
    cookie = request.COOKIES.get(ninja_simple_jwt_settings.JWT_REFRESH_COOKIE_NAME)
    if cookie is None:
        raise AuthenticationError()
    access_token, _ = get_access_token_from_refresh_token(cookie)
    return {"access": access_token}



from ninja import Schema


class SignInRequest(Schema):
    username: str
    password: str


class MobileSignInResponse(Schema):
    refresh: str
    access: str


class MobileTokenRefreshRequest(Schema):
    refresh: str


class MobileTokenRefreshResponse(Schema):
    access: str


class WebSignInResponse(Schema):
    access: str



from typing import Any

from django.core.management.base import BaseCommand

from ninja_simple_jwt.jwt.key_creation import make_and_save_key_pair


class Command(BaseCommand):
    help = "Create RSA key pair."

    def handle(self, *args: Any, **kwargs: Any) -> None:
        private_key_path, public_key_path = make_and_save_key_pair()
        print(f"Key pair created: \n {private_key_path}\n {public_key_path}")
